ERROR_PLATFORM_NOT_SUPPORTED = '''
Platform not supported.
If you are using Alexa, you must pass the request trough the Alexa Skills Kit.
If you are using Google Assistant you must use an API gateway and create a route named /googleAssistantDialogflowV1
If you are using Bixby you must use an API gateway and create a route named /samsungBixbyV1
If you are using Siri you must use an API gateway and create a route named /appleSiriV1
As of your current framework version, no other platform is supported.
'''
